#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <cctype>
#include <sstream>

#include <errno.h>
#include <string.h>

#include <unistd.h>
#include <signal.h>

using namespace std;

constexpr int MAX_WIDTH = 100;
constexpr int MAX_HEIGHT = 100;

constexpr int RUN_PRESENTATION_ERR = 4;
constexpr int RUN_WRONG_ANSWER_ERR = 5;
constexpr int RUN_CHECK_FAILED = 6;

constexpr char ON = '*';

constexpr char MAX_TOKEN_LENGTH = 32;

int other_pid = 0;

[[noreturn]]
void do_check_failed(void)
{
    close(1);
    // drain input
    int c;
    while ((c = getchar()) != EOF) {}
    if (other_pid > 1) kill(other_pid, SIGINT);
    exit(RUN_CHECK_FAILED);
}

[[noreturn]]
void do_presentation_error(void)
{
    close(1);
    // drain input
    int c;
    while ((c = getchar()) != EOF) {}
    if (other_pid > 1) kill(other_pid, SIGINT);
    exit(RUN_PRESENTATION_ERR);
}

[[noreturn]]
void do_wrong_answer(void)
{
    close(1);
    // drain input
    int c;
    while ((c = getchar()) != EOF) {}
    if (other_pid > 1) kill(other_pid, SIGINT);
    exit(RUN_WRONG_ANSWER_ERR);
}

[[noreturn]]
void do_ok(void)
{
    close(1);
    // drain input
    int c;
    while ((c = getchar()) != EOF) {}
    exit(0);
}

int safe_read_int(int *curc, int *retval)
{
    while (isspace(*curc)) *curc = getchar();
    if (*curc == EOF) return EOF;

    char buf[MAX_TOKEN_LENGTH + 8];
    int len = 0;
    while (*curc != EOF && !isspace(*curc)) {
        if (len == MAX_TOKEN_LENGTH) {
            cerr << "input token length exceeds " << MAX_TOKEN_LENGTH << endl;
            do_presentation_error();
        }
        buf[len++] = *curc;
        *curc = getchar();
    }
    buf[len] = 0;

    errno = 0;
    char *eptr = NULL;
    long val = strtol(buf, &eptr, 10);
    if (errno || *eptr || eptr == buf || (int) val != val) {
        cerr << "input token '" << buf << "' is not int" << endl;
        do_presentation_error();
    }
    *retval = val;
    return 0;
}

/*
EJUDGE_USER_ID
EJUDGE_USER_LOGIN
EJUDGE_USER_NAME
EJUDGE_CONTEST_ID
EJUDGE_RUN_ID
EJUDGE_TEST_NUM
 */

/*
[1]: test input
[2]: interactor output (to checker)
[3]: correct answer
[4]: pid
*/

int main(int argc, char *argv[])
{
    if (argc != 5) {
        cerr << "wrong number of args" << endl;
        do_check_failed();
    }

    int user_id = -1;
    if (const char *s = getenv("EJUDGE_USER_ID"); s) {
        user_id = stol(s);
    }
    if (user_id <= 0 || user_id > 100000000) {
        cerr << "EJUDGE_USER_ID environment is not set" << endl;
        do_check_failed();
    }

    const char *user_login = getenv("EJUDGE_USER_LOGIN");
    if (!user_login) {
        cerr << "EJUDGE_USER_LOGIN environment is not set" << endl;
        do_check_failed();
    }

    /*
    const char *user_name = getenv("EJUDGE_USER_NAME");
    if (!user_name) {
        cerr << "EJUDGE_USER_NAME environment is not set" << endl;
        do_check_failed();
    }
    */

    int contest_id = -1;
    if (const char *s = getenv("EJUDGE_CONTEST_ID"); s) {
        contest_id = stol(s);
    }
    if (contest_id <= 0 || contest_id > 100000000) {
        cerr << "EJUDGE_CONTEST_ID environment is not set" << endl;
        do_check_failed();
    }

    int run_id = -1;
    if (const char *s = getenv("EJUDGE_RUN_ID"); s) {
        run_id = stol(s);
    }
    if (run_id <= 0 || run_id > 100000000) {
        cerr << "EJUDGE_RUN_ID environment is not set" << endl;
        do_check_failed();
    }

    int test_num = -1;
    if (const char *s = getenv("EJUDGE_TEST_NUM"); s) {
        test_num = stol(s);
    }
    if (test_num <= 0 || test_num > 100000000) {
        cerr << "EJUDGE_TEST_NUM environment is not set" << endl;
        do_check_failed();
    }

    fprintf(stderr, "User: %d, %s\n", user_id, user_login);
    fprintf(stderr, "Contest: %d\n", contest_id);
    fprintf(stderr, "Run: %d\n", run_id);
    fprintf(stderr, "Test: %d\n", test_num);

    const char *out_path = argv[2];

    other_pid = strtol(argv[4], NULL, 10);
    //fprintf(stderr, "pid: %d\n", other_pid);

    ifstream inp;

    signal(SIGPIPE, SIG_IGN);

    /*
    for (int i = 0; i < argc; ++i) {
        fprintf(stderr, "[%d]: %s\n", i, argv[i]);
    }
    */

    inp.open(argv[1]);
    if (!inp) {
        cerr << "cannot open input file " << argv[1] << endl;
        do_check_failed();
    }

    bool do_not_reply = false;

    vector<string> matr;
    {
        string s;
        while (getline(inp, s)) {
            string::size_type len = s.length();
            while (len > 0 && isspace(s[len - 1])) --len;
            s.resize(len);
            matr.push_back(move(s));
        }
    }

    {
        auto len = matr.size();
        while (len > 0 && !matr[len - 1].length()) --len;
        matr.resize(len);
    }

    const size_t height = matr.size();

    if (!height) {
        cerr << "input matrix is empty";
        do_check_failed();
    }
    if (height > MAX_HEIGHT) {
        cerr << "input matrix is too big";
        do_check_failed();
    }
    const size_t width = matr[0].length();
    for (auto &s : matr) {
        if (s.length() != width) {
            cerr << "input matrix is not rectagonal";
            do_check_failed();
        }
        for (auto &c : s) {
            if (c != '.') c = ON;
        }
    }
    cerr << "input matrix: height: " << height << ", width: " << width << endl;
    for (auto &s : matr) {
        cerr << s << endl;
    }

    {
        ostringstream oss;
        oss << height << "\n" << width << "\n";
        for (size_t row = 0; row < height; ++row) {
            vector<int> runs;
            size_t col = 0;
            while (col < matr[row].length()) {
                int cnt = 0;
                while (col < width && matr[row][col] == ON) {
                    ++col; ++cnt;
                }
                if (cnt > 0) {
                    runs.push_back(cnt);
                }
                while (col < width && matr[row][col] != ON) {
                    ++col;
                }
            }
            oss << runs.size();
            for (auto r : runs) {
                oss << " " << r;
            }
            oss << "\n";
        }
        for (size_t col = 0; col < width; ++col) {
            vector<int> runs;
            size_t row = 0;
            while (row < height) {
                int cnt = 0;
                while (row < height && matr[row][col] == ON) {
                    ++row; ++cnt;
                }
                if (cnt > 0) {
                    runs.push_back(cnt);
                }
                while (row < height && matr[row][col] != ON) {
                    ++row;
                }
            }
            oss << runs.size();
            for (auto r : runs) {
                oss << " " << r;
            }
            oss << "\n";
        }

        string os = oss.str();
        const char *ptr = os.c_str();
        size_t z = os.length();

        if (height <= 20 && width <= 20) {
            cerr << "Encoded matrix: " << endl;
            cerr << os << endl;
        }

        if (!z) {
            cerr << "no initial data" << endl;
            do_check_failed();
        }
        while (z) {
            ssize_t r = write(1, ptr, z);
            if (r <= 0) {
                cerr << "write error to pipe: " << strerror(errno) << endl;
                do_presentation_error();
            }
            ptr += r;
            z -= r;
        }
    }

    // f*** C++ with vector<bool>!
    vector<vector<int> > visited(height);
    for (auto &v : visited)
        v.resize(width);

    int coord_count = 0;
    int query_count = 0;
    int c = getchar();
    while (1) {
        int row{}, col{}, q{};

        if (safe_read_int(&c, &row) < 0) break;
        if (safe_read_int(&c, &col) < 0) {
            cerr << "unexpected EOF reading column" << endl;
            do_presentation_error();
        }
        if (safe_read_int(&c, &q) < 0) {
            cerr << "unexpected EOF reading value" << endl;
            do_presentation_error();
        }
        if (row < 0 || row >= int(height)) {
            cerr << "row " << row << " is out of range" << endl;
            do_presentation_error();
        }
        if (col < 0 || col >= int(width)) {
            cerr << "column " << col << " is out of range" << endl;
            do_presentation_error();
        }
        if (visited[row][col]) {
            cerr << "cell (" << row << "," << col << ") was already requested or reported" << endl;
            do_presentation_error();
        }
        if (q == 0) {
            if (matr[row][col] == ON) {
                cerr << "wrong answer (" << row << "," << col << ") IS FILLED" << endl;
                do_wrong_answer();
            }
            ++coord_count;
            visited[row][col] = true;
        } else if (q == 1) {
            if (matr[row][col] != ON) {
                cerr << "wrong answer (" << row << "," << col << ") IS NOT FILLED" << endl;
                do_wrong_answer();
            }
            ++coord_count;
            visited[row][col] = true;
        } else if (q == -1) {
            ++query_count;
            ++coord_count;
            visited[row][col] = true;

            ostringstream oss;
            int val = 0;
            if (matr[row][col] == ON) val = 1;
            oss << val << "\n";

            string os = oss.str();
            const char *ptr = os.c_str();
            size_t z = os.length();

            if (!z) {
                cerr << "no answer" << endl;
                do_check_failed();
            }
            if (!do_not_reply) {
                while (z) {
                    ssize_t r = write(1, ptr, z);
                    if (r < 0 && errno == EPIPE) {
                        do_not_reply = true;
                        break;
                    } else if (r <= 0) {
                        cerr << "write error to pipe: " << strerror(errno) << endl;
                        do_presentation_error();
                    }
                    ptr += r;
                    z -= r;
                }
            }
        } else {
            cerr << "invalid value " << q << endl;
            do_presentation_error();
        }
    }

    if (coord_count != int(height * width)) {
        cerr << "only " << coord_count << " positions reported instead of " << (height * width) << endl;
        do_presentation_error();
    }

    cerr << "QUERY: " << query_count << endl;

    // write to DB
    if (0) {
        ostringstream oss;
        ostringstream osq;

        osq << "INSERT INTO nonograms SET user_id = " << user_id;
        osq << ", contest_id = " << contest_id;
        osq << ", test_num = " << test_num;
        osq << ", user_login = \'" << user_login << "\'";
        osq << ", best_query = " << query_count;
        osq << " ON DUPLICATE KEY UPDATE best_query = LEAST(best_query, " << query_count << ");";
        auto oq = osq.str();

        oss << "mysql --user ejudge -D ejudge --password=53ieYw4HnH4SLBGD";
        oss << " -e \"" << oq << "\"";

        fprintf(stderr, "%s\n", oq.c_str());
        auto os = oss.str();
        system(os.c_str());
    }

    int score = 0;
    if (query_count == 0) {
        score = 5;
    } else if (query_count >= int(width) * int(height)) {
        score = 1;
    } else {
        int grp = (width * height - 1) / 3;
        score = 4 - (query_count - 1)/ grp;
    }

    cerr << "SCORE: " << score << endl;

    {
        ofstream outf(out_path);
        outf << score << endl;
    }
    do_ok();
}
